/*
 boards/MMC5.c
  the ExROM boards

 * MacFCEU 0.8a1
 *
 * Copyright notice for this file:
 *  mmc5.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* None of this code should use any of the iNES bank switching wrappers. */

#include "mapinc.h"
#include "MMC5.h"
#include "../cheat.h"

static void (*sfunc)(int P);
static void (*psfun)(void);

void MMC5RunSound(int32 Count);
void MMC5RunSoundHQ(void);

//static INLINE void MMC5SPRVROM_BANK1(uint32 A, uint32 V)
//static INLINE void MMC5BGVROM_BANK1(uint32 A,uint32 V)
#define MMC5_BANK1(w,AB,V) do{uint32 _V=V; \
	if(!*CHRptr) break; \
	_V&=*CHRmask1; \
	w[AB]=&CHRptr[0][_V<<10]-(AB<<10); \
} while(0)

#define MMC5_BANK2(w,AB,V) do{uint32 _V=V; \
	if(!*CHRptr) break; \
	_V&=*CHRmask2; \
	w[AB]=w[AB+1]=&CHRptr[0][_V<<11]-(AB<<10); \
} while(0)

#define MMC5_BANK4(w,AB,V)	do{ uint32 _V=V; \
	if(!*CHRptr) break; \
	_V&=*CHRmask4; \
	w[AB]=w[AB+1]=w[AB+2]=w[AB+3]=&CHRptr[0][_V<<12]-(AB<<10); \
} while(0)

#define MMC5_BANK8(w,V)		do{ uint32 _V=V; \
	if(!*CHRptr) break; \
	_V&=*CHRmask8; \
	*w=w[1]=w[2]=w[3]=w[4]=w[5]=w[6]=w[7]=&CHRptr[0][_V<<13]; \
} while(0)

static uint8	PRGBanks[4];
static uint16	CHRBanksA[8], CHRBanksB[4];

static uint8	WRAMPage;
static uint8	WRAMMaskEnable[2];

uint8		MMC5ExMode;
uint8		MMC5ABMode;		/* A=0, B=1 */
uint8		MMC5CHRHigh = 0;	/* 5130h */

/* these were in ppu.c */
uint8		MMC5Ex;
uint32		MMC5VROMMask;
uint8		*MMC5VROMPtr;
uint8		MMC5SplitCtrl;
uint8		MMC5SplitY;
uint8		MMC5SplitBank;

static uint8	IRQScanline, IRQEnable;
static uint8	NTAMirroring, NTFill, ATFill;

static uint8	MMC5IRQR;
#define kIRQInFrame		0x40
#define kIRQLineHit		0x80

static uint8	MMC5LineCounter;
static uint8	MMC5psize, MMC5vsize;
static uint8	mul[3];

static uint8	*WRAM = 0;
static uint8	*MMC5fill = 0;
uint8		*MMC5ExRAM = 0;

static uint32	MMC5WRAMsize;
static uint8	MMC5WRAMIndex[8];

static uint8	MMC5ROMWrProtect[4];
static uint8	MMC5MemIn[5];

static void MMC5CHRA(void);
static void MMC5CHRB(void);

typedef struct cartdata
{
	uint32	crc32;
	uint8	size;
} cartdata;


// ETROM seems to have 16KB of WRAM, ELROM seems to have 8KB
// EWROM seems to have 32KB of WRAM

#define MMC5_NOCARTS		14
cartdata	MMC5CartList[MMC5_NOCARTS] =
{
	{0x9c18762b,2},		/* L'Empereur */
	{0x26533405,2},
	{0x6396b988,2},

	{0xaca15643,2},		/* Uncharted Waters */
	{0xfe3488d1,2},		/* Dai Koukai Jidai */

	{0x15fe6d0f,2},		/* BKAC             */
	{0x39f2ce4b,2},		/* Suikoden              */

	{0x8ce478db,2},		/* Nobunaga's Ambition 2 */
	{0xeee9a682,2},

	{0x1ced086f,2},		/* Ishin no Arashi */

	{0xf540677b,4},		/* Nobunaga...Bushou Fuuun Roku */

	{0x6f4e4312,4},		/* Aoki Ookami..Genchou */

	{0xf011e490,4},		/* Romance of the 3 Kingdoms 2 */
	{0x184c2124,4},		/* Sangokushi 2 */
};


int DetectMMC5WRAMSize(uint32 crc32);
int DetectMMC5WRAMSize(uint32 crc32)
{
	int		x = 0;

	if (MMC5_NOCARTS>0)
		do
			if (crc32==MMC5CartList[x].crc32)
			{
				FCEU_printf(
" >8KB external WRAM present.  Use UNIF if you hack the ROM image.\n");
					return MMC5CartList[x].size*8;
			}
		while(++x<MMC5_NOCARTS);

	return 8;
}

static void BuildWRAMSizeTable(void)
{
	int		x = 0;

	do
		switch(MMC5WRAMsize/(8*1024))
		{
		case 0: MMC5WRAMIndex[x] = 255; break;
		case 1: MMC5WRAMIndex[x] = x>3 ? 255:0; break;
		case 2: MMC5WRAMIndex[x] = x>>2&1; break;
		case 4: MMC5WRAMIndex[x] = x>3 ? 255:x&3; break;
		}
	while(++x<8);
}

static void MMC5CHRA(void)
{
	int		x;

	switch(MMC5vsize&3)
	{
	case 0:
		setchr8(CHRBanksA[7]);
		MMC5_BANK8(MMC5SPRVPage,CHRBanksA[7]);
		break;
	case 1:
		setchr4(0x0000,CHRBanksA[3]);
		setchr4(0x1000,CHRBanksA[7]);
		MMC5_BANK4(MMC5SPRVPage,0,CHRBanksA[3]);
		MMC5_BANK4(MMC5SPRVPage,4,CHRBanksA[7]);
		break;
	case 2:
		setchr2(0x0000,CHRBanksA[1]);
		setchr2(0x0800,CHRBanksA[3]);
		setchr2(0x1000,CHRBanksA[5]);
		setchr2(0x1800,CHRBanksA[7]);
		MMC5_BANK2(MMC5SPRVPage,0,CHRBanksA[1]);
		MMC5_BANK2(MMC5SPRVPage,2,CHRBanksA[3]);
		MMC5_BANK2(MMC5SPRVPage,4,CHRBanksA[5]);
		MMC5_BANK2(MMC5SPRVPage,6,CHRBanksA[7]);
		break;
	case 3:
		x = 0;
		do
		{
			setchr1(x<<10,CHRBanksA[x]);
			MMC5_BANK1(MMC5SPRVPage,x,CHRBanksA[x]);
		} while(++x<8);
	}
}

static void MMC5CHRB(void)
{
	int		x;

	switch(MMC5vsize&3)
	{
	case 0:
		setchr8(CHRBanksB[3]);
		MMC5_BANK8(MMC5BGVPage,CHRBanksB[3]);
		break;
	case 1:
		setchr4(0x0000,CHRBanksB[3]);
		setchr4(0x1000,CHRBanksB[3]);
		MMC5_BANK4(MMC5BGVPage,0,CHRBanksB[3]);
		MMC5_BANK4(MMC5BGVPage,4,CHRBanksB[3]);
		break;
	case 2:
		setchr2(0x0000,CHRBanksB[1]);
		setchr2(0x0800,CHRBanksB[3]);
		setchr2(0x1000,CHRBanksB[1]);
		setchr2(0x1800,CHRBanksB[3]);
		MMC5_BANK2(MMC5BGVPage,0,CHRBanksB[1]);
		MMC5_BANK2(MMC5BGVPage,2,CHRBanksB[3]);
		MMC5_BANK2(MMC5BGVPage,4,CHRBanksB[1]);
		MMC5_BANK2(MMC5BGVPage,6,CHRBanksB[3]);
		break;
	case 3:
		x = 0;
		do
		{
			setchr1(x<<10,CHRBanksB[x&3]);
			MMC5_BANK1(MMC5BGVPage,x,CHRBanksB[x&3]);
		} while(++x<8);
	}
}

static void FASTAPASS(2) MMC5WRAM(uint32 A, uint32 V)
{
	//printf("%02x\n",V);
	if ((V=MMC5WRAMIndex[V&7])!=255)
	{
		setprg8r(0x10,A,V);
		MMC5MemIn[A-0x6000>>13] = 1;
	} else
		MMC5MemIn[A-0x6000>>13] = 0;
}

static void MMC5PRG(void)
{
	int		x;

	switch(MMC5psize&3)
	{
	case 0:
		MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] =
			MMC5ROMWrProtect[2] = MMC5ROMWrProtect[3] = 1;
		setprg32(0x8000,((PRGBanks[1]&0x7F)>>2));
		MMC5MemIn[1] = MMC5MemIn[2] = MMC5MemIn[3] = MMC5MemIn[4] = 1;
		break;
	case 1:
		if (PRGBanks[1]&0x80)
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 1;
			setprg16(0x8000,(PRGBanks[1]>>1));
			MMC5MemIn[1] = MMC5MemIn[2] = 1;
		} else
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 0;
			MMC5WRAM(0x8000,PRGBanks[1]&7&0xFE);
			MMC5WRAM(0xA000,(PRGBanks[1]&7&0xFE)+1);
		}

		MMC5MemIn[3] = MMC5MemIn[4] =
			MMC5ROMWrProtect[2] = MMC5ROMWrProtect[3] = 1;
		setprg16(0xC000,PRGBanks[3]>>1&0x3F);
		break;
	case 2:
		if (PRGBanks[1]&0x80)
		{
			MMC5MemIn[1] = MMC5MemIn[2] = 1;
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 1;
			setprg16(0x8000,PRGBanks[1]>>1&0x3F);
		} else
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 0;
			MMC5WRAM(0x8000,PRGBanks[1]&7&0xFE);
			MMC5WRAM(0xA000,(PRGBanks[1]&7&0xFE)+1);
		}

		if (PRGBanks[2]&0x80)
		{
			MMC5ROMWrProtect[2] = MMC5MemIn[3] = 1;
			setprg8(0xC000,PRGBanks[2]&0x7F);
		} else
		{
			MMC5ROMWrProtect[2] = 0;
			MMC5WRAM(0xC000,PRGBanks[2]&7);
		}

		MMC5MemIn[4] = MMC5ROMWrProtect[3] = 1;
		setprg8(0xE000,PRGBanks[3]&0x7F);
		break;
	case 3:
		x = 0;
		do
			if (PRGBanks[x]&0x80)
			{
				MMC5ROMWrProtect[x] = 1;
				setprg8(0x8000+(x<<13),PRGBanks[x]&0x7F);
				MMC5MemIn[1+x] = 1;
			} else
			{
				MMC5ROMWrProtect[x] = 0;
				MMC5WRAM(0x8000+(x<<13),PRGBanks[x]&7);
			}
		while(++x<3);

		MMC5MemIn[4] = MMC5ROMWrProtect[3] = 1;
		setprg8(0xE000,PRGBanks[3]&0x7F);
	}
}

static DECLFW(MMC5W5100)
{
	uint32		t;
	int		x;

	switch(A)
	{
	case 0x5100:
		MMC5psize = V;
		MMC5PRG();
		break;
	case 0x5101:
		MMC5vsize = V;
		if (!MMC5ABMode)
			{MMC5CHRB(); MMC5CHRA();}
		else
			{MMC5CHRA(); MMC5CHRB();}
		break;
	case 0x5102: WRAMMaskEnable[0] = V; break;
	case 0x5103: WRAMMaskEnable[1] = V; break;
	case 0x5104: MMC5ExMode = V&=3; break;
	case 0x5105:
		x = t = 0;
		do
			switch((V>>x+x)&3)
			{
			case 0: PPUNTARAM|= t; vnapage[x] = NTARAM; break;
			case 1: PPUNTARAM|= t; vnapage[x] = NTARAM+0x400; break;
			case 2: PPUNTARAM|= t; vnapage[x] = MMC5ExRAM; break;
			case 3: PPUNTARAM&=~t; vnapage[x] = MMC5fill; break;
			}
		while(t+=t,++x<4);
		NTAMirroring = V;
		break;
	case 0x5106:
		if (V!=NTFill)
		{
			t = V, t|= t<<8, t|= t<<16;
			FCEU_dwmemset(MMC5fill,t,0x3C0);
		}
		NTFill = V;
		break;
	case 0x5107:
		if (V!=ATFill)
		{
			t = V|(V<<2)|(V<<4)|(V<<6), t|= t<<8, t|= t<<16;
			FCEU_dwmemset(MMC5fill+0x3C0,t,0x40);
		}
		ATFill = V;
		break;
	case 0x5108:
	case 0x5109:
	default:
		break;
	case 0x5113:
		WRAMPage = V;
		MMC5WRAM(0x6000,V&=7);
		break;
	case 0x5114:
	case 0x5115:
	case 0x5116:
	case 0x5117:
		PRGBanks[A&=3] = V;
		MMC5PRG();
		break;
	case 0x5120:
	case 0x5121:
	case 0x5122:
	case 0x5123:
	case 0x5124:
	case 0x5125:
	case 0x5126:
	case 0x5127:
		MMC5ABMode = 0;
		CHRBanksA[A] = V|(MMC5CHRHigh<<8&0x300);
		MMC5CHRA();
		break;
	case 0x5128:
	case 0x5129:
	case 0x512A:
	case 0x512B:
		MMC5ABMode = 1;
		CHRBanksB[A&=3] = V;
		MMC5CHRB();
		break;
	case 0x5130:
		MMC5CHRHigh = V;
	}
}

static DECLFR(MMC5R5200)
{
	uint8		r = A>>8;

	if ((A-=0x5204)==0)
	{
		r = MMC5IRQR;
		if (!fceuindbg)
		{
			X6502_IRQEnd(FCEU_IQEXT); //this was outside `!fceuindbg`
			MMC5IRQR&= kIRQInFrame;
		}
	} else if (--A==0) //5205h
		return mul[2]=mul[0]*mul[1];
	else if (--A==0) //5206h
		return mul[2]=(mul[0]*mul[1])>>8;

	return r; //XFrame.DB
}

static DECLFW(MMC5W5200)
{
	if ((A-=0x5200)==0)
		MMC5SplitCtrl = V;
	else if (--A==0) //1
		MMC5SplitY = V>>3&0x1F;
	else if (--A==0) //2
		MMC5SplitBank = V&=0x3F;
	else if (--A==0) //3
		{IRQScanline = V; goto clr;}
	else if (--A==0) //4
	{
		if (!(IRQEnable=V&0x80)) //only if bit #7 is zero; 2010.04.09
		{
		clr:
			X6502_IRQEnd(FCEU_IQEXT);
		}
	} else
		mul[A-1] = V;
}

static DECLFR(MMC5R5C00)
{
	if (MMC5ExRAM && MMC5ExMode&2)
		return MMC5ExRAM[A-0x5C00];
	else
		return A>>8;
}

static DECLFW(MMC5W5C00)
{
	if (MMC5ExMode==3) /* ROM */
		return;
	if (MMC5ExMode!=2) /* others call for clearing the pixel when out of frame */
		if (!(MMC5IRQR&kIRQInFrame))
			V = 0;
	if (MMC5ExRAM)
		MMC5ExRAM[A-0x5C00] = V;
}

static DECLFR(MMC5R6000)
{
	if (MMC5MemIn[(A-0x6000)>>13])
		return Page[A>>11][A];
	else
		return A>>8; //XFrame.DB
}

static DECLFW(MMC5W6000)
{
	if (A>=0x8000)
		if (MMC5ROMWrProtect[A-0x8000>>13])
			return;
	if (MMC5MemIn[A-0x6000>>13])
		if (((WRAMMaskEnable[0]&3)|((WRAMMaskEnable[1]&3)<<2))==6)
			Page[A>>11][A] = V;
}

void MMC5Synco(void);
void MMC5Synco(void)
{
	uint32		t;
	int		x;

	MMC5PRG();
	x = t = 0;
	do /* mirroring for MMC5 is actually pretty complicated */
		switch((NTAMirroring>>x+x)&3)
		{
		case 0: PPUNTARAM|= t, vnapage[x] = NTARAM; break;
		case 1: PPUNTARAM|= t, vnapage[x] = NTARAM+0x400; break;
		case 2: PPUNTARAM|= t, vnapage[x] = MMC5ExRAM; break;
		case 3: PPUNTARAM&=~t, vnapage[x] = MMC5fill; break;
		}
	while(t+=t,++x<4);

	MMC5WRAM(0x6000,WRAMPage&7);
	if (!MMC5ABMode)
	{
		MMC5CHRB();
		MMC5CHRA();
	} else
	{
		MMC5CHRA();
		MMC5CHRB();
	}

	t = NTFill|(NTFill<<8);
	t|= t<<16;
	FCEU_dwmemset(MMC5fill,t,0x3c0);

	t = ATFill|(ATFill<<2)|(ATFill<<4)|(ATFill<<6);
	t|= t<<8;
	t|= t<<16;
	FCEU_dwmemset(MMC5fill+0x3c0,t,0x40);

	X6502_IRQEnd(FCEU_IQEXT);
}

void MMC5_hb(int scanline)
{
	if (scanline==240)
	{
		MMC5LineCounter = 0;
		MMC5IRQR = kIRQInFrame;
		return;
	}
	if (MMC5LineCounter<240)
	{
		if (MMC5LineCounter==IRQScanline)
		{
			MMC5IRQR|= kIRQLineHit;
			if (IRQEnable&0x80)
				X6502_IRQBegin(FCEU_IQEXT);
		}
		++MMC5LineCounter;
	} else if (MMC5LineCounter==240)
		MMC5IRQR = 0;
}

static uint8	CHRBanksAL[8], CHRBanksBL[4];

void MMC5Restore(int version);
void MMC5Restore(int version)
{
	if (version<9814)
	{
		int	i;

		i = 0;
		do
			CHRBanksA[i] = CHRBanksAL[i];
		while(++i<8);
		i = 0;
		do
			CHRBanksB[i] = CHRBanksBL[i];
		while(++i<4);
	}
	MMC5Synco();
}

typedef struct MMC5APU/*(38)*/
{
	uint16	wl[2];
	uint8	env[2];
	uint8	enable;
	uint8	running;
	uint8	raw;
	uint8	rawcontrol;

	int32	dcount[2];
	int32	vcount[2];

	int32	BC[3];
} MMC5APU;

static MMC5APU	MMC5Sound;


static void Do5PCM()
{
	int32		x, start, end;
	uint32		V;

	start = MMC5Sound.BC[2];
	if ((end=(SOUNDTS<<16)/soundtsinc)<=start)
		return;
	MMC5Sound.BC[2] = end;

	if (!(MMC5Sound.rawcontrol&0x40) && (V=MMC5Sound.raw*2)!=0)
	{
		x = start;
		do
			FCEUWave[x>>4]+= V;
		while(++x<end);
	}
}

static void Do5PCMHQ()
{
	int32		x;
	uint32		V;

	if (!(MMC5Sound.rawcontrol&0x40) && (x=MMC5Sound.BC[2])<SOUNDTS
			&& (V=MMC5Sound.raw<<5)!=0)
		do
			FCEUWaveHi[x]+= V;
		while(++x<SOUNDTS);
	MMC5Sound.BC[2] = SOUNDTS;
}

/* added; 2010.04.09 */
static DECLFR(MMC5R5015)
{
(void)A;
	return (MMC5Sound.wl[0]!=0)+(MMC5Sound.wl[1]!=0)*2;
}

static DECLFW(MMC5W5000)
{
	uint32		Ads2 = (A&=0x1F)>>2;

	GameExpSound.Fill[SCMMC5] = MMC5RunSound;
	GameExpSound.HiFill[SCMMC5] = MMC5RunSoundHQ;

	switch(A)
	{
	case 0:
	case 4:
		if (sfunc)
			sfunc(Ads2);
		MMC5Sound.env[Ads2] = V;
		break;
	case 2:
	case 6:
		if (sfunc)
			sfunc(Ads2);
		MMC5Sound.wl[Ads2]&=~0xFF;
		MMC5Sound.wl[Ads2]|= V;
		break;
	case 3:
	case 7://printf("%04x:$%02x\n",A,V>>3);
		MMC5Sound.wl[Ads2]&=~0x700;
		MMC5Sound.wl[Ads2]|= V<<8&0x700;
		MMC5Sound.running|= 1<<Ads2;
		break;
	case 0x10:
		if (psfun)
			psfun();
		MMC5Sound.rawcontrol = V;
		break;
	case 0x11:
		if (psfun)
			psfun();
		MMC5Sound.raw = V;
		break;
	case 0x15:
		if (sfunc)
			{sfunc(0); sfunc(1);}
		MMC5Sound.running&= V;
		MMC5Sound.enable = V;
		//printf("%02x\n",V);
	}
}

static void Do5SQ(int P)
{
	static int	tal[4] = {1,2,4,6};

	int32		wl;
	int32		x, end;

	x = MMC5Sound.BC[P];
	if ((end=(SOUNDTS<<16)/soundtsinc)<=x)
		return;
	MMC5Sound.BC[P] = end;

	if ((wl=((uint32)MMC5Sound.wl[P]+1)<<18)>=(8L<<18)
			&& (MMC5Sound.running&(P+1)))
	{
		int32	amp = MMC5Sound.env[P]<<4&0xF0;
		int32	rthresh = tal[MMC5Sound.env[P]>>6];
		int	dc = MMC5Sound.dcount[P];
		int	vc = MMC5Sound.vcount[P];

		do
		{
			if (dc<rthresh)
				FCEUWave[x>>4]+= amp;
			if ((vc-=nesincsize)<=0)
				do
					dc = (dc+1)&7;
				while((vc+=wl)<=0);
		} while(++x<end);

		MMC5Sound.dcount[P] = dc;
		MMC5Sound.vcount[P] = vc;
	}
}

static void Do5SQHQ(int P)
{
	static int	tal[4] = {1,2,4,6};

	int32		wl;

	if ((wl=(MMC5Sound.wl[P]+1)*2)>=8*2
			&& (MMC5Sound.running&(P+1)))
	{
		int32	amp = MMC5Sound.env[P]<<8&0xF00;
		int32	rthresh = tal[MMC5Sound.env[P]>>6];
		int	dc = MMC5Sound.dcount[P];
		int	vc = MMC5Sound.vcount[P];
		int	x = MMC5Sound.BC[P];

		do
		{
			if (dc<rthresh)
				FCEUWaveHi[x]+= amp;
			if (--vc<=0)	/* Less than zero when first started. */
				vc = wl, dc = (dc+1)&7;
		} while(++x<SOUNDTS);
		MMC5Sound.dcount[P] = dc;
		MMC5Sound.vcount[P] = vc;
	}

	MMC5Sound.BC[P] = SOUNDTS;
}

void MMC5RunSoundHQ(void)
{
	Do5SQHQ(0);
	Do5SQHQ(1);
	Do5PCMHQ();
}

void MMC5HiSync(int32 ts)
{
	MMC5Sound.BC[0] = MMC5Sound.BC[1] = MMC5Sound.BC[2] = ts;
}

void MMC5RunSound(int32 Count)
{
	Do5SQ(0);
	Do5SQ(1);
	Do5PCM();
	MMC5Sound.BC[0] = MMC5Sound.BC[1] = MMC5Sound.BC[2] = Count;
}

void MMC5_ESI()
{
	GameExpSound.RChange[SCMMC5] = MMC5_ESI;
	if (FSettings.SndRate)
		if (FSettings.soundq)
			sfunc = Do5SQHQ, psfun = Do5PCMHQ;
		else
			sfunc = Do5SQ, psfun = Do5PCM;
	else
		sfunc = 0, psfun = 0;

	memset(MMC5Sound.BC,0,sizeof(MMC5Sound.BC));
	memset(MMC5Sound.vcount,0,sizeof(MMC5Sound.vcount));
	GameExpSound.HiSync[SCMMC5] = MMC5HiSync;
}

void NSFMMC5_Init()
{
	memset(&MMC5Sound,0,sizeof(MMC5Sound));
	mul[0] = mul[1] = 0;

	if ((MMC5ExRAM=(uint8*)FCEU_gmalloc(1024))==0)
		return;

	MMC5_ESI();
	SetReadHandler(0x5205,0x5206,MMC5R5200);
	SetReadHandler(0x5C00,0x5FEF,MMC5R5C00);
	SetWriteHandler(0x5000,0x5015,MMC5W5000);
	SetWriteHandler(0x5205,0x5206,MMC5W5200);
	SetWriteHandler(0x5C00,0x5FEF,MMC5W5C00);
	MMC5ExMode = 2;
}

void GenMMC5Close()
{
	MMC5Ex = 0; /* was MMC5Hack in fceu.c */
	if (MMC5ExRAM)
		{FCEU_free(MMC5ExRAM); MMC5ExRAM = 0;}
	/* leak fixes; 2010.04.09 */
	if (MMC5fill)
		{FCEU_free(MMC5fill); MMC5fill = 0;}
	if (WRAM)
		{FCEU_free(WRAM); WRAM = 0;}
}

void GenMMC5Power()
{
	int		x;

	x = 0; do PRGBanks[x] =~0; while(++x<4);
	x = 0; do CHRBanksA[x] = 0xFF; while(++x<8);
	x = 0; do CHRBanksB[x] = 0xFF; while(++x<4);
	WRAMMaskEnable[0] = WRAMMaskEnable[1] = ~0;

	MMC5psize = MMC5vsize = 3;
	MMC5ExMode = 0;

	NTAMirroring = NTFill = ATFill = 0xFF;

	MMC5Synco();

	SetReadHandler(0x5015,0x5015,MMC5R5015);
	SetReadHandler(0x5204,0x5206,MMC5R5200);
	SetReadHandler(0x5C00,0x5FFF,MMC5R5C00);
	SetReadHandler(0x6000,0xFFFF,MMC5R6000);

	SetWriteHandler(0x5000,0x5015,MMC5W5000);
	SetWriteHandler(0x5100,0x5130,MMC5W5100);
	SetWriteHandler(0x5200,0x5206,MMC5W5200);
	SetWriteHandler(0x5C00,0x5FFF,MMC5W5C00);
	SetWriteHandler(0x6000,0xFFFF,MMC5W6000);

	//GameHBIRQHook=MMC5_hb; -- MMC5_hb called in PPU.c
	FCEU_CheatSetRAM(8,0x6000,WRAM);
	FCEU_CheatSetRAM(1,0x5C00,MMC5ExRAM);
}

static SFORMAT MMC5StateRegs[] =
{
	{ PRGBanks, 4, "PRGB"},
	{ CHRBanksAL, 8, "CHRA"}, //CHRx are the old ones, updated
	{ CHRBanksBL, 4, "CHRB"},
	{ &CHRBanksA[0], sizeof(CHRBanksA[0])|FCEUSTATE_RLSB, "CHA0"}, //0.5.4 was bad at this
	{ &CHRBanksA[1], sizeof(CHRBanksA[1])|FCEUSTATE_RLSB, "CHA1"},
	{ &CHRBanksA[2], sizeof(CHRBanksA[2])|FCEUSTATE_RLSB, "CHA2"},
	{ &CHRBanksA[3], sizeof(CHRBanksA[3])|FCEUSTATE_RLSB, "CHA3"},
	{ &CHRBanksA[4], sizeof(CHRBanksA[4])|FCEUSTATE_RLSB, "CHA4"},
	{ &CHRBanksA[5], sizeof(CHRBanksA[5])|FCEUSTATE_RLSB, "CHA5"},
	{ &CHRBanksA[6], sizeof(CHRBanksA[6])|FCEUSTATE_RLSB, "CHA6"},
	{ &CHRBanksA[7], sizeof(CHRBanksA[7])|FCEUSTATE_RLSB, "CHA7"},
	{ &CHRBanksB[0], sizeof(CHRBanksB[0])|FCEUSTATE_RLSB, "CHB0"},
	{ &CHRBanksB[1], sizeof(CHRBanksB[1])|FCEUSTATE_RLSB, "CHB1"},
	{ &CHRBanksB[2], sizeof(CHRBanksB[2])|FCEUSTATE_RLSB, "CHB2"},
	{ &CHRBanksB[3], sizeof(CHRBanksB[3])|FCEUSTATE_RLSB, "CHB3"},
	{ &WRAMPage, 1, "WRMP"},
	{ WRAMMaskEnable, 2, "WRME"},
	{ &MMC5ABMode, 1, "ABMD"},
	{ &MMC5CHRHigh, 1, "5130"},
	{ &IRQScanline, 1, "IRQS"},
	{ &IRQEnable, 1, "IRQE"},
	{ &MMC5ExMode, 1, "CHRM"},
	{ &NTAMirroring, 1, "NTAM"},
	{ &NTFill, 1, "NTFL"},
	{ &ATFill, 1, "ATFL"},

	{ &MMC5Sound.wl[0], 2|FCEUSTATE_RLSB, "SDW0"},
	{ &MMC5Sound.wl[1], 2|FCEUSTATE_RLSB, "SDW1"},
	{ MMC5Sound.env, 2, "SDEV"},
	{ &MMC5Sound.enable, 1, "SDEN"},
	{ &MMC5Sound.running, 1, "SDRU"},
	{ &MMC5Sound.raw, 1, "SDRW"},
	{ &MMC5Sound.rawcontrol, 1, "SDRC"},
//	{ &MMC5Sound.dcount[0], 4|FCEUSTATE_RLSB, "DCN0"},
//	{ &MMC5Sound.dcount[0], 4|FCEUSTATE_RLSB, "DCN1"},
//	{ &MMC5Sound.vcount[0], 4|FCEUSTATE_RLSB, "VCN0"},
//	{ &MMC5Sound.vcount[0], 4|FCEUSTATE_RLSB, "VCN1"},

	{ &MMC5SplitCtrl,1,"SPLM"},
	{ &MMC5SplitY,1,"SPLS"},
	{ &MMC5SplitBank,1,"SPLP"},
	{0}
};

void GenMMC5_Init(CartInfo *info,
	int wsize,
	int battery)
{
	if ((MMC5fill=(uint8*)FCEU_gmalloc(1024))==0
			|| (MMC5ExRAM=(uint8*)FCEU_gmalloc(1024))==0)
	{
	fail:
		GenMMC5Close();
		return;
	}

	if ((MMC5WRAMsize=wsize*=1024)!=0)
	{
		if ((WRAM=(uint8*)FCEU_gmalloc(wsize))==0)
			goto fail;
		SetupCartPRGMapping(0x10,WRAM,wsize,1);
		AddExState(WRAM,wsize,0,"WRAM");
	}
	BuildWRAMSizeTable();

	if (battery)
	{
		info->SaveGame[0] = WRAM;
		if (wsize<=16*1024)
			info->SaveGameLen[0] = wsize ? 8192:0;
		else
			info->SaveGameLen[0] = 32768;
	}

	info->Power = GenMMC5Power;
	info->Close = GenMMC5Close;
	GameStateRestore = MMC5Restore;

	MMC5Ex = 1;
	MMC5VROMMask = CHRmask4[0];
	MMC5VROMPtr = CHRptr[0];
	MMC5SplitCtrl = MMC5SplitY = MMC5SplitBank = 0;

	MMC5_ESI();

	AddExState(MMC5StateRegs,~0,0,0);
	AddExState(WRAM,wsize,0,"WRAM");
	AddExState(MMC5ExRAM,1024,0,"ERAM");
}

void ExROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,DetectMMC5WRAMSize(info->CRC32),info->battery);
}

// ELROM seems to have 0KB of WRAM
// EKROM seems to have 8KB of WRAM
// ETROM seems to have 16KB of WRAM
// EWROM seems to have 32KB of WRAM
// ETROM and EWROM are battery-backed, EKROM isn't.

void ETROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,16,info->battery);
}

void ELROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,0,0);
}

void EWROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,32,info->battery);
}

void EKROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,8,info->battery);
}
