/*
 utility/asm6502.c

 * MacFCEU 0.8
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 TODO
	( (indirect)
	incbin
	pad
	opt
	mem
	dc.b
	dc.w
	dcb
	dcw
	index (mapper)
	code
	segment (ignore)
	byte (explicit header)
	zp
	zeropage
	- (temporary labels)
	% (binary)
-	.shl (<<), .shr (>>)
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "types.h"
#include "../interp/interpreter.h"

#include "asm6502.h"
#include "disasm6502.h"

static const char *IDBreak(const char *str,
	const char *e)
{
	const char	*s = str;

	while(s<e)
		switch(*s)
		{
		case '@':
		case '_':
		case '.':
		CASE_NUMERIC
		CASE_ALPHABETIC
		next:	++s;
			break;

		default:
#ifdef CHR_NB
			if (*s==CHR_NB)
				goto next;
#endif
			return s;
		}
	return s;
}

enum
{ /* corresponding index is +OPN_COUNT */

	_block		= OPN_COUNT,
	_db,
	_dw,
	_end,
	_equ,
	_fill,
	_include,
	_module,
	_opt,
	_org,
	_p02,
	_processor
};

const char	*keyw[] =
{
	"BLOCK",
	"DB",
	"DW",
	"END",
	"EQU",
	"FILL",
	"INCLUDE",
	"MODULE",
	"P02",
	"PROCESSOR",
	"OPT",
	"ORG"
};

/* returns ID offset (past hash), non-positive on error */
int32 X6502_AsmGetID(const char *s,
	const char *e,
	uint8 *n,
	const _asm6502frame *f)
{
	const uint8	*l;
	uint8		h, z;

	if (!n || !f || !s || e<=s)
		return 0;
	if ((l=f->lab)==0)
		return -1;
	*n = toupper(*s++)&0x3F, z = h = 0;
	while(s<e)
		if (!++z)
			return 0;
		else
			h+= n[z] = toupper(*s++);
	if ((h=h<<6|*n)==0) /* '@' is valid */
		h = 0x80; /* hash can't be null */
	*n++ = h;
	/* hash and type match */
	for(; *l++; l+=*l,l+=1+4)
		if (*l==z && h==l[-1] && !memcmp(l+1,n,z))
			return l-f->lab;
	return f->lab-l; /* becomes ~total */
}

/* returns ID offset, or <0 on error */
int X6502_AsmAddID(const char *s,
	const char *e,
	uint8 *n,
	_asm6502frame *f)
{
	
	uint8		*d;
	int32		x;

	if ((x=X6502_AsmGetID(s,e,n,f))>=0)
		return !x ? _e_identifier_error:x;
	if ((d=realloc(f->lab,(e-s)+1+4-x))==0)
		return _e_memory_error;
	f->lab = d, d-= x;
	d[-1] = *n++;
	*d++ = x = e-s-1;
	memcpy(d,n,x);
	*(d+=x) = 0;

	return d-f->lab;
}

static int getNameIndex(const char *s,
	const char *e)
{
	char		n[10];
	int		i, z = e-s;
	uint8		m = 0;

	if (*s=='.')
		++s, m = 1;
	if ((z=e-s)>9)
		goto nope;
	i = 0;
	do
		n[i] = toupper(*s++);
	while(++i,s<e);

	i = 0;
	if (z==3 && !m)
	{
		const char	*t = *x6502_opnametable;

		do
			if (*t==*n && t[1]==n[1] && t[2]==n[2])
				return i;
		while(t+=3,++i<OPN_COUNT);
	} else
		i = OPN_COUNT;
#ifdef KEYW_REQUIRES_DOT
	if (m)
#endif
	{ /* keyword (e.g., .equ) */
		const char	**t = keyw;

		--z;
		do
			if (**t==*n && !memcmp(t,n+1,z))
				return i;
		while(++t,++i<OPN_COUNT+GET_ARRAY_COUNT(keyw));
	}
nope:
	return _e_identifier_error;
}

/* returns an intermediate value - adjust in parsing */
static int getMode(const char *s,
	const char *e)
{
	e = parse_rskipwsp(e,s);
	if (s>=e)
		return OPM_IP;
	if (*s=='#')
		return OPM_IM;
	if (*s=='(')
		return OPM_IN;
	return OPM_AB;
}

static int getInstruction(int t)
{
	int		i = 0;

	do
		if ((x6502_opn_table[i]&(OPB_BAD-1))==t)
			return i;
	while(++i<256);

	return -1;
}

static const char *getv(const char *s,
	const char *e,
	evalvalue *i,
	void *u)
{
	uint8		b[256];
	const char	*n;
	int32		x;

	if (s>=e)
		return NULL;
	n = IDBreak(s,e);
	if (*s=='*')
	{
		if (i)
			*i = ((const _asm6502frame*)u)->org;
		return s+1;
	}
	if ((x=X6502_AsmGetID(s,n,b,u))<=0)
		return NULL;
	e = ((const char*)((const _asm6502frame*)u)->lab)+x+(n-s);

	if (i)
		*i = *e|(e[1]<<8);

	return s;
}

int X6502_Assemble(const char *s,
	int (*writef)(uint32,
		uint8),
	_asm6502frame *f)
{
	uint8		b[256];
	_asm6502frame	af;
	const char	*e;
	int		r, m;
	uint8		c = 0;

	if (s==0)
		return -1;
	r = 0;
	if ((e=strlen(s)+s)==s)
		return 0; /* empty */
	if (f==0)
	{ /* clear */
		f = &af;
		f->org = 0, f->flg = 0;
		f->lab = 0;
	}

	while(s<e)
	{
		evalvalue	v;
		const char	*n;
		long		l;
		int		x;

		if (*s=='\n' || *s=='\r') /* new line */
		{
			f->flg&=~_IntoLine, ++s;
			continue;
		}
		if (c)
			continue;
		if (*(n=parse_skipwsp(s,e))==';')
			{c = 1, s = n+1; continue;}
		if (n!=s)
			f->flg|= _IntoLine;
		m =-1;
		if (*(n=IDBreak(s,e))==':')
			goto labo;
		if ((x=getNameIndex(s,n))>=0)
			goto gotn;
		if (!(f->flg&_IntoLine))
		{ /* identifier: equate/label */
		labo:	if ((r=X6502_AsmAddID(s,n,b,f))!=0)
			{ /* test label rescan ignore option */
				if (r!=_e_identifier_error || !(f->opt&_Relabel))
					break;
			} else
				m = r+n-s, r = 0;
			if (*n==':')
				++n;
		}
		if (*(n=parse_skipwsp(n,e))==';')
			{c = 1, s = n+1; continue;}
		if ((n=IDBreak(n,e))<=s)
		{
			if (*s=='=') /* '=' equate */
				{++n, x = _equ; goto gotn;}
			r =-1; /* syntax error */
			break;
		}
		/* macro or instruction */
		if ((x=getNameIndex(s,n))<0)
			{r = x; break;} /* can't identify */
		/* keyword, possible equate, etc. */
	gotn:	s = n, r = 0;
		switch(x)
		{
		case _block: /* asign block, skip */
			l = evaluate_line(s,e,&v,&r,getv,NULL,f,NULL);
			if ((unsigned)r>=(-1u>>1))
				goto end;
			r = v, v = f->org;
			s+= x;
			x = IMAX(x);
			goto lset;
		case _db:
		case _dw:
			break;
		case _end:
			s = e, f->flg = _EndLine;
			break;
		case _equ: /* set label to explicit value */
			l = evaluate_line(s,e,&v,&r,getv,NULL,f,NULL);
			if ((unsigned)r>=(-1u>>1))
				goto end;
			x = IMAX(x);
			goto lset;
		case _fill: /* fill bytes of length, optional value */
			l = evaluate_line(s,e,&v,&r,getv,NULL,f,NULL);
			if ((unsigned)r>=(-1u>>1))
				goto end;
			x = v, v = 0;
			if (*(n=parse_skipwsp(s+l,e))==',')
			{
				s = parse_skipwsp(n+1,e);
				l = evaluate_line(s,e,&v,&r,getv,NULL,f,NULL);
				if ((unsigned)r>=(-1u>>1))
					goto end;
			}
			if (!writef)
				f->org+= x;
			else
				do
					writef(f->org,v);
				while(++f->org,--x);
			x = IMAX(x);
			goto lset;
		case _include: /* read this file in place of line */
			break;
		case _module: /* switch section type */
			break;
		case _opt:
			
			break;
		case _org: /* set section position */
			l = evaluate_line(s,e,&v,&r,getv,NULL,f,NULL);
			if ((unsigned)r>=(-1u>>1))
				goto end;
			s+= l;
			/* there is also an optional fill value... */
			f->org = v, r = 0, x = IMAX(x);

		default: /* label at org */
		lset:	if (m>=0)
			{
				uint8	*d;

				if ((d=f->lab)==NULL)
					{r = _e_read_error; break;}
				d+= m, *d = v, d[1] = v>>8;
			}
			f->org+= r;
		}
		s = parse_skipwsp(n,e);
		if (s>=e)
			break;
		if (x>=OPN_COUNT)
			continue; /* not encoding an instruction */
		m = getMode(s,e);

		/* parameters */
		l = evaluate_line(s,e,&v,&r,getv,NULL,f,NULL);
		if (m==OPM_AB && v<256)
			m = OPM_ZP; /* optimize size */
		if ((unsigned)(x=getInstruction(x|(m<<8)))>255)
			if ((m==OPM_ZP || m==OPM_AB)
					&& (unsigned)(x=getInstruction(x|(OPM_RL<<8)))<256)
			{
				v-= f->org, --v;
				/* check relative range */
				if (v!=(int8)v)
					{r = _e_range_error; break;}
			} else /* last resort: check ZP with only ABS available */
				if (m!=OPM_ZP || (unsigned)(x=getInstruction(x|(OPM_AB<<8)))>255)
					{r =-1; break;} /* no such instruction */
		if (writef)
			writef(f->org,v);
		++f->org;
		if ((unsigned)(m-OPM_IB)<=OPM_AY-OPM_IB)
		{
			if (writef)
				writef(f->org,v>>8);
			++f->org;
		}
	}
end:	if (f==&af && f->lab)
		free(af.lab);
	return r;
}

#ifdef SAMPLE
uint8		*obuf;

int bwrite(uint32 a, /* can be NULL (check syntax) */
	uint8 v)
{
	if (obuf)
		{obuf[a] = v; return 1;}
	return 0;
}

char *err_strs[] =
{
	"syntax error",
	"memory error",
	"read error",
	"write error",
	"incomplete expression",
	"interrogator already open",
	"no such identifier",
	"handler not supplied",
	"stack over/underflow",
	"relative out of range"
};

static int asm6502(FILE *f)
{
	_asm6502frame	af;
	char		*s;
	uint32		z, l, x = 1;

	memset(&af,0,sizeof(_asm6502frame));

	s = malloc(0); /* needed for first-empty line(s) case */
	while(z=~0,l=x,f && (s=freadline(f,&z,s,NULL,&x,'\\'))!=NULL)
	{
		const char	*e = s+z;
		long		r;

		if ((r=X6502_Assemble(parse_skipwsp(s,e),bwrite,&af))!=0)
		{
			if (r==_e_param_separation)
				r = _e_general_error;
			if ((unsigned)r>=(-1u>>1) && (r=~r)<GET_ARRAY_COUNT(err_strs))
				fprintf(stderr,
"### %s\n"
"### % *^\n"
"### %s\n",
					s,r,err_strs[r]);
			goto end;
		}
		if (af.flg==_EndLine)
			break;
		af.flg&=~_IntoLine;
	}
	if (z)
		z = _e_memory_error;
end:	free(s);
	if (af.lab)
		free(af.lab);
	return z;
}

int main(int argc,
	char **argv)
{
	FILE		*f = NULL;
	const char	*n = "asm.out";
	char		**a, *s;
	long		m = 1u<<16;
	int		r =-1;

	if (argc<=0 || (a=argv)==NULL)
		return 0;
	if (!--argc || (s=*++a)==NULL)
	{
		fprintf(stderr,"### usage: %s filename\n",
			*argv);
		fputs(
"#\n"
"#  -            use stdin\n"
"#  -O           set output name\n"
"#\n",
			stderr);
		return 1;
	}
	do
		if (*s=='-')
		{
			if (!s[1])
				if (f)
				{
					fprintf(stderr,"### %s - cant stdin *and* open file.\n",
						*argv,s);
					goto bail;
				} else
					f = stdin, s = "dev::stdin";
			else if (s[1]=='E')
				n = s+2;
			else
			{
				fprintf(stderr,"### %s - unknown option '%s'.\n",
					*argv,s);
				return 1;
			}
		} else if (f)
		{
			fprintf(stderr,"### %s - one file at a time.\n",
				*argv,s);
			goto bail;
		} else
			f = fopen(s,"r");
	while(--argc);

	if (!f)
	{
		fprintf(stderr,"### %s - could not get/open '%s'.\n",
			*argv,s);
		return 1;
	}

	obuf = calloc(1,m);
	if ((r=asm6502(f))==0)
		if ((f=fopen(n,"w"))==NULL)
			fprintf(stderr,"### %s - could not open '%s' for writing.\n",
				*argv,n);
		else
		{
			if (fwrite(obuf,1,m,f)!=m)
				r = _e_write_error;
			fclose(f);
		}
	free(obuf);
	obuf = NULL;
bail:	if (f && f!=stdin)
		fclose(f);
	if (r)
		fprintf(stderr,"### %s - fatal error (%d).\n",
			*argv,r), r = 1;
	return r;
}
#endif
